clc; clear all; close all; format compact;
%{
Digital twin designed by Paul A. Valle: paul.valle@tectijuana.edu.mx
https://biomath.xyz/
This algorithm can be applied to fit the numerical values of parameters 
rho1, rho8, rho9, and to determine results concerning the statistics and 
goodness of fit.

Journal: Entropy.
Manuscript: Mechanistic modelling of biomass growth, glucose consumption and
ethanol production by Kluyveromyces marxianus in batch fermentation.
Authors: Salazar, Y.; Valle, P.A.; Rodriguez, E.; Soto, N.O.; Paez, J.B.; 
and Reyes, F.J.
Year: 2023

Experimental data is extracted from a 'xlslx' file with three sheets. Where
sheet 1 contains the measured data for Biomass, sheet 2 for Glucose, and 
sheet 3 for Ethanol. Each sheet contains 1 + 17*2 columns, the first column
corresponds to the time, then every two columns contain the data for each 
strain since the fermentation was made in duplicate and samples were taken
each hour from t = 0 to t = 13, i.e., 14*2 data points for each variable of 
every strain.

Experimental data for Biomass [x(t)]:
---------------------------------------------------------------------------
t(:,1) x1_1(:,2) x1_2(:,3) x2_1(:,4) x2_2(:,5) ... x17_1(:,34) x17_2(:,35)
---------------------------------------------------------------------------

Experimental data for Glucose [y(t)]:
---------------------------------------------------------------------------
t(:,1) y1_1(:,2) y1_2(:,3) y2_1(:,4) y2_2(:,5) ... y17_1(:,34) y17_2(:,35)
---------------------------------------------------------------------------

Experimental data for Ethanol [z(t)]:
---------------------------------------------------------------------------
t(:,1) z1_1(:,2) z1_2(:,3) z2_1(:,4) z2_2(:,5) ... z17_1(:,34) z17_2(:,35)
---------------------------------------------------------------------------
Experimental data may be requested to: nsoto@itdurango.edu.mx and
jpaez@itdurango.edu.mx
%}

X = readmatrix("K_marxianus.xlsx",'Sheet',1); % Biomass experimental data
Y = readmatrix("K_marxianus.xlsx",'Sheet',2); % Glucose experimental data
Z = readmatrix("K_marxianus.xlsx",'Sheet',3); % Ethanol experimental data
n = (size(X,2)-1)/2; % Number of strains
t = X(:,1); % Time array from the measurements: t = B(:,1) = G(:,1) = E(:,1)

p2 = 2.2814;        % Affinity with substrate constant (rho_2)
p3 = 1.0658E-3;     % Inhibition rate of biomass growth due to product accumulation (rho_3)
p4 = 7.2751E-3;     % Biomass death rate (rho_4)
p5 = 56.8928E-3;    % Consumption rate for biomass growth (rho_5)
p6 = 71.8424E-3;    % Consumption rate for ethanol production (rho_6)
p7 = 824.2332E-9;   % Glucose spontaneous decomposition rate (rho_7)
p10 = 149.8985E-3;  % Ethanol degradation rate (rho_10)
P = [p2;p3;p4;p5;p6;p7;p10]; % Fixed parameters array
p0 = [1E-1;1E-1;1E-1]; % Initial values for the lsqcurvefit function: rho_1, rho_8, rho_9 

%{
BioMath function
--- Inputs ---
t: time array from the measurements, size(t) = [14 1]
X: Experimental data of Biomass, size(X) = [14 35]
Y: Experimental data of Glucose, size(Y) = [14 35]
Z: Experimental data of Ethanol, size(Z) = [14 35]
p0: Initial values for the lsqcurvefit function
P: Fixed parameters values
n: Number of strains (17)
--- Outputs ---
rho: Fitted values for rho_1, rho_8, rho_9, size(rho) = [17 3]
SE: Standard error of the fitted values, size(SE) = [17 3]
CI95: 95% confidence intervals for the fitted values, size(CI95) = [17 6]
R2: R squared results. R2(:,1) for the complete set of data,  R2(:,2) for 
biomass, R2(:,3) for glucose, and R2(:,4) for ethanol, size(R2) = [17 4]
RSS: Residual sum of squares for the complete set of data, size(RSS) = [17 1]
AIC: Akaike Information Criterion results for each strain, size(AIC) = [17 1]
xa: Approximated data for biomass in all strains, size(xa) = [14 17]
ya: Approximated data for glucose in all strains, size(ya) = [14 17]
za: Approximated data for ethanol in all strains, size(za) = [14 17]
%}
[rho,SE,CI95,R2,RSS,AIC,xa,ya,za] = BioMath(t,X,Y,Z,p0,P,n);

% Table of results to be displayed in the Command Window
rho1 = rho(:,1); rho8=rho(:,2); rho9=rho(:,3);
R2_phi = R2(:,1); R2_Biomass = R2(:,2); R2_Glucose = R2(:,3); R2_Ethanol = R2(:,4);
SE_rho1 = SE(:,1); SE_rho8 = SE(:,2); SE_rho9 = SE(:,3);
CI95_rho1 = CI95(:,1:2); CI95_rho8 = CI95(:,3:4); CI95_rho9 = CI95(:,5:6);
Strain = (1:n)';
Results = table(Strain,rho1,SE_rho1,CI95_rho1,rho8,SE_rho8,CI95_rho8,...
    rho9,SE_rho9,CI95_rho9,RSS,AIC,R2_phi,R2_Biomass,R2_Glucose,R2_Ethanol);
fprintf('\n\n Results for the K. marxianus fitting process: \n')
disp(Results)

% Saving results in a xlsx file
writetable(Results,'Results.xlsx') % Fitted values and statistics
writematrix([t,xa],'K_marxianus_approximated.xlsx','Sheet',1) % Biomass approximated data
writematrix([t,ya],'K_marxianus_approximated.xlsx','Sheet',2) % Glucose approximated data
writematrix([t,za],'K_marxianus_approximated.xlsx','Sheet',3) % Ethanol approximated data
%% BioMath function
function [rho,SE,CI95,R2,RSS,AIC,xa,ya,za] = BioMath(t,X,Y,Z,p0,P,n)
    rho = zeros(n,length(p0));
    SE = zeros(n,length(p0));
    CI95 = zeros(n,length(p0)*2);
    R2 = zeros(n,4);
    RSS = zeros(n,1);
    AIC = zeros(n,1);
    xa = zeros(length(t),n);
    ya = zeros(length(t),n);
    za = zeros(length(t),n);

    function sys = Digital_Twin(p0,t) % Digital twin function for the K. marxianus strains
        %--- Inputs ---
        % p0: Initial guess of parameters to be fitted
        % t: Time array of the experimental data
        %--- Outputs ---
        % sys: Approximated values for Biomass, Glucose, and Ethanol

        dt = 1E-5;                      % Delta_t for Euler's method
        time = (0:dt:max(t))';          % Time array
        N = round(max(t)/dt);           % Number of iterations
        x = zeros(N+1,1); x(1) = x0;    % Biomass array initialization
        y = zeros(N+1,1); y(1) = y0;    % Glucose array initialization
        z = zeros(N+1,1); z(1) = z0;    % Ethanol array initialization
        
        % Parameters:
        p1 = p0(1); p8 = p0(2); p9 = p0(3); % Parameters to be fitted
        p2 = P(1); p3 = P(2); p4 = P(3); p5 = P(4); p6 = P(5); p7 = P(6); p10 = P(7); % Fixed parameters
        
        % Computational model with Euler's method
        for i=1:N  
            x(i+1) = x(i) + (+p1*x(i)*y(i)/(p2+y(i)) -p3*x(i)*z(i) -p4*x(i))*dt; % Biomass
            y(i+1) = y(i) + (-p5*x(i)*y(i) -p6*y(i)*z(i) -p7*y(i))*dt;           % Glucose   
            z(i+1) = z(i) + (+p8*x(i)*z(i) +p9*y(i)*z(i) -p10*z(i))*dt;          % Ethanol
        end
        
        % Sampling to match the experimental data array
        xk = zeros(length(t),1); % Array for the sampled data of Biomass
        yk = zeros(length(t),1); % Array for the sampled data of Glucose
        zk = zeros(length(t),1); % Array for the sampled data of Ethanol
        for i = 1:length(t)
            k = abs(time-t(i)) < 1E-9;
            xk(i) = x(k); % Data sampled from Biomass x(t)
            yk(i) = y(k); % Data sampled from Glucose y(t)
            zk(i) = z(k); % Data sampled from Ethanol z(t)
        end
        
        sys = [xk;yk;zk];
    end
    
    function [R2,RSS,AIC] = GOF(fe,fa,K) % Goodness of fit (GOF) function
        %--- Inputs ---
        % fe: Array of experimental data
        % fa: Array of approximated data
        %  K: Total number de parameters of the model
        %--- Outputs ---
        %  R2: R squared value
        % RSS: Residual Sum of Squares value
        % AIC: Akaike Information Criterion value

        n_data = length(fe); % Total number of observations
        RSS = 0; TSS = 0;
        for i = 1:n_data
            RSS = RSS + (fe(i)-fa(i))^2;    % Residual sum of squares
            TSS = TSS + (fe(i)-mean(fe))^2; % Total sum of squares
        end
        R2 = 1 - RSS/TSS;
        AIC = n_data*log(RSS/n_data)+2*K+(2*K*(K+1))/(n_data-K-1);
    end

    function plot_results(t,xe,ye,ze,xa,ya,za,s) % Illustrating results
        set(figure(s),'color','w')
        set(gcf,'units','centimeters','position',[2 2 17 22])
        
        subplot(3,1,1) % Biomass dynamics [x(t)] in g/L
        grid on; hold on; box on;
        plot(t,xe,'x','linewidth',1,'color',[0.3,0.5,0.2])
        plot(t,xa(:,s),'-','linewidth',1,'color',[0.05,0.05,0.4])
        L=legend('$x_e(t)$','$x_a(t)$','location','southeast');
        set(L,'interpreter','latex','fontsize',11)
        xlim([0,14]); xticks(0:1:14)
        ylim([0,4.5]); yticks(0:0.5:4.5)
        xlabel('$t$ $[hours]$','interpreter','latex','fontsize',11)
        ylabel('$x(t):Biomass$ $[g/L]$','interpreter','latex','fontsize',11)
        set(gca,'fontname','times new roman','fontsize',11)
        
        subplot(3,1,2) % Glucose dynamics [y(t)] in g/L
        grid on; hold on; box on;
        plot(t,ye,'x','linewidth',1,'color',[0.3,0.5,0.2])
        plot(t,ya(:,s),'-','linewidth',1,'color',[0.05,0.05,0.4])
        L=legend('$y_e(t)$','$y_a(t)$','location','northeast');
        set(L,'interpreter','latex','fontsize',11)
        xlim([0,14]); xticks(0:1:14)
        ylim([-24*0.07,24]); yticks(0:3:24)
        xlabel('$t$ $[hours]$','interpreter','latex','fontsize',11)
        ylabel('$y(t):Glucose$ $[g/L]$','interpreter','latex','fontsize',11)
        set(gca,'fontname','times new roman','fontsize',11)
        
        subplot(3,1,3) % Ethanol dynamics [z(t)] in g/L
        grid on; hold on; box on;
        plot(t,ze,'x','linewidth',1,'color',[0.3,0.5,0.2])
        plot(t,za(:,s),'-','linewidth',1,'color',[0.05,0.05,0.4])
        L=legend('$z_e(t)$','$z_a(t)$','location','southeast');
        set(L,'interpreter','latex','fontsize',11)
        xlim([0,14]); xticks(0:1:14)
        ylim([-10*0.07,10]); yticks(0:1:10)
        xlabel('$t$ $[hours]$','interpreter','latex','fontsize',11)
        ylabel('$z(t):Ethanol$ $[g/L]$','interpreter','latex','fontsize',11)
        set(gca,'fontname','times new roman','fontsize',11)
        
        sg=sgtitle(['{\it K. marxianus} strain: ',num2str(s)]); 
        set(sg,'fontname','times new roman','fontsize',12)
        
        % Saving the in silico experimentation as pdf, png, or jpeg format
        %exportgraphics(gcf,['strain_',num2str(s),'.pdf'],'ContentType','vector')
        %exportgraphics(gcf,['strain_',num2str(s),'.png'],'ContentType','image','Resolution',600)
        %print(['strain_',num2str(s),'.jpeg'],'-djpeg','-r600')
        % Remove comments if you want to save the Figures.
    end
    
    % Analysis for each strain
    for s = 1:n
        x1 = X(:,2*s);   y1 = Y(:,2*s);   z1 = Z(:,2*s);    % First measurement
        x2 = X(:,2*s+1); y2 = Y(:,2*s+1); z2 = Z(:,2*s+1);  % Second measurement 
        xe = mean([x1,x2],2); ye = mean([y1,y2],2); ze = mean([z1,z2],2);  % Average value 
        x0 = xe(1); y0 = ye(1); z0 = ze(1); % Initial conditions
        fe = [xe;ye;ze]; % Array of experimental data as a column vector

        %{
        lsqcurvefit function
        --- Inputs ---
        Digital_Twin: K. Marxianus computational model
          p0: Parameters to be fitted, rho_1, rho_8, rho_9
           t: Time array of the experimental data
          fe: Experimental data
         low: Lower bounds of fitting parameters 
         sup: Upper bounds of fitting parameters
        opts: optimotions of the lsqcurvefit function
        --- Outputs ---
        estimate: Fitted values of parameters
        residual: Residuals matrix = sys - fe
        jacobian: Jacobian at the solution, returned as a real matrix
        %}
        low = -ones(length(p0),1); sup = ones(length(p0),1);
        opts = optimoptions('lsqcurvefit','MaxFunctionEvaluations',1E3,...
        'MaxIterations',1E3,'FunctionTolerance',1E-9);
        

        [estimate,~,residual,~,~,~,jacobian] = lsqcurvefit(@Digital_Twin,p0,t,fe,low,sup,opts);
        rho(s,:) = estimate;
       

        % Approximated data with the fitted parameters
        fa = Digital_Twin(estimate,t); xyz = reshape(fa,[],3);
        xa(:,s) = xyz(:,1); ya(:,s) = xyz(:,2); za(:,s) = xyz(:,3);
        plot_results(t,xe,ye,ze,xa,ya,za,s) % Illustrating results

        % Statistics: standard errors and 95% confidence intervals for each fitted parameter value
        dof = numel(fe) - length(p0); % Degrees of freedom
        covv = full(inv(jacobian'*jacobian))*((residual'*residual)/dof); % Variance-Covariance matrix
        se = sqrt(diag(covv)); SE(s,:) = se'; % Standard error
        tv = 2.0227; % Two-tail t-value with a significance level (alpha) of 0.05 and dof
        % See http://www.ttable.org/
        ci95 = zeros(length(p0),2); % Confidence intervals initialization matrix
        for j = 1:length(p0)
            ci95(j,1) = estimate(j) - se(j)*tv; % Lower limit of the 95% CI
            ci95(j,2) = estimate(j) + se(j)*tv; % Upper limit of the 95% CI
            CI95(s,2*j-1) = ci95(j,1); CI95(s,2*j) = ci95(j,2);
        end
       
        %{
        Table of results for each strain
        parameter: Name of fitted parameters
         estimate: Fitted values of parameters
               se: Standard Error of the fitted value
             ci95: 95% Confidence Interval of the fitted value
        %}
        fprintf(['\n Results for K. marxianus strain: ',num2str(s),'\n'])
        parameter = ['rho1';'rho8';'rho9'];
        results = table(parameter,estimate,se,ci95); disp(results)

        % Goodness of fit for strain: 's' 
        K = length(p0)+length(P); % Total number of parameters
        [r2,rss,aic] = GOF(fe,fa,K);   R2(s,1) = r2; RSS(s,1) = rss; AIC(s,1) = aic;
        [r2x,~,~] = GOF(xe,xa(:,s),K); R2(s,2) = r2x;
        [r2y,~,~] = GOF(ye,ya(:,s),K); R2(s,3) = r2y;
        [r2z,~,~] = GOF(ze,za(:,s),K); R2(s,4) = r2z;
    end
end